#include "bs_blob_collective.h"
#include "bs_mem.h"
#include "bs_pointmass.h"

#include <math.h>

static bs_int32_t 
bs_blob_collective_get_unique_id()
{
  static bs_int32_t i = 0; 

  i++; 

  return i;   
}

typedef struct bs_blob_container_st
{
  bs_blob_t *pBlob;
  bs_int32_t id; 
} bs_blob_container_t; 

static bs_blob_container_t* 
bs_blob_container_create(bs_blob_t *pBlob)
{
  bs_blob_container_t *pBlobContainer; 
  
  BS_MEM_ALLOC_TYPE(pBlobContainer, bs_blob_container_t, 1); 
  pBlobContainer->pBlob = pBlob; 
  pBlobContainer->id = bs_blob_collective_get_unique_id(); 
  
  return pBlobContainer; 
}

static void 
bs_blob_container_destructor(void *pUserData)
{
  bs_blob_container_t *pBlobContainer; 
  
  BS_ASSERT(pUserData != NULL); 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  bs_blob_destroy(pBlobContainer->pBlob); 
  BS_MEM_FREE(pBlobContainer);   
}

typedef struct bs_joint_container_st
{
  bs_joint_t *pJoint; 
  bs_int32_t idA, idB; 
} bs_joint_container_t; 

static bs_joint_container_t* 
bs_joint_container_create(bs_joint_t *pJoint, 
                          bs_int32_t idA, 
                          bs_int32_t idB)
{
  bs_joint_container_t *pJointContainer; 

  BS_MEM_ALLOC_TYPE(pJointContainer, bs_joint_container_t, 1); 
  pJointContainer->pJoint = pJoint; 
  pJointContainer->idA = idA; 
  pJointContainer->idB = idB; 
  
  return pJointContainer; 
}

static void 
bs_joint_container_destroy(void *pUserData)
{
  bs_joint_container_t *pJointContainer;   
  
  BS_ASSERT(pUserData != NULL); 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  bs_joint_destroy(pJointContainer->pJoint); 
  BS_MEM_FREE(pJointContainer); 
}

bs_blob_collective_t* 
bs_blob_collective_create(bs_float_t xPos, 
                          bs_float_t yPos, 
                          bs_list_factory_t *pFactory)
{
  bs_blob_collective_t *pCollective; 
  bs_blob_container_t *pBlobContainer; 

  BS_MEM_ALLOC_TYPE(pCollective, bs_blob_collective_t, 1); 

  pCollective->pFactory = pFactory; 
  
  pCollective->pBlobs = bs_list_factory_get_list(pFactory); 
  
  pBlobContainer = bs_blob_container_create(bs_blob_create(xPos, yPos, 0.4f)); 
  bs_list_add_element_to_front(pCollective->pBlobs, (void*)pBlobContainer, 
    bs_blob_container_destructor); 
  
  pCollective->pJoints = bs_list_factory_get_list(pFactory); 
  
  return pCollective; 
}

void 
bs_blob_collective_destroy(bs_blob_collective_t *pCollective)
{
  BS_ASSERT(pCollective != NULL);
  BS_ASSERT(pCollective->pBlobs != NULL); 
  
  bs_list_factory_give_list(pCollective->pFactory, pCollective->pBlobs); 
}

typedef struct bs_blob_collective_find_largest_st
{
  bs_float_t maxRadius; 
  bs_blob_t *pMotherBlob; 
  bs_int32_t id; 
} bs_blob_collective_find_largest_t; 

static void 
bs_blob_collective_find_largest_cb(void *pUserData, 
                                   void *pAccumulator)
{
  bs_blob_container_t *pBlobContainer; 
  bs_blob_collective_find_largest_t *pFindData;   
  bs_float_t radius; 
  
  BS_ASSERT(pUserData != NULL); 
  BS_ASSERT(pAccumulator != NULL); 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  pFindData = (bs_blob_collective_find_largest_t*) pAccumulator; 
  
  radius = bs_blob_get_radius(pBlobContainer->pBlob); 
  if(radius > pFindData->maxRadius)
  {
    pFindData->maxRadius = radius; 
    pFindData->pMotherBlob = pBlobContainer->pBlob;    
    pFindData->id = pBlobContainer->id;     
  }
}

static void 
bs_blob_collective_scale_old_joints(void *pUserData, 
                                    bs_float_t scaleFactor, 
                                    void *pAccumulator)
{
  bs_joint_container_t *pJointContainer; 
  bs_blob_collective_find_largest_t *pFindData;   
  
  BS_ASSERT(pUserData != NULL); 
  BS_ASSERT(pAccumulator != NULL); 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  pFindData = (bs_blob_collective_find_largest_t*) pAccumulator; 

  if(pFindData->id == pJointContainer->idA || 
     pFindData->id == pJointContainer->idB)
  {
    bs_joint_scale(pJointContainer->pJoint, scaleFactor); 
  }    
}

typedef struct bs_blob_collective_create_new_joints_st
{
  bs_blob_collective_t *pCollective; 
  bs_blob_container_t *pNewBlobContainer; 
} bs_blob_collective_create_new_joints_t; 

static void 
bs_blob_collective_create_new_joints_cb(void *pUserData, 
                                        void *pAccumulator)
{
  bs_blob_container_t *pListBlob; 
  bs_blob_collective_create_new_joints_t *pCreateData; 
  bs_pointmass_t *pListBlobPointmass, *pNewBlobPointmass; 
  bs_joint_t *pNewJoint; 
  bs_joint_container_t *pJointContainer; 
  bs_float_t dist; 
  
  BS_ASSERT(pUserData != NULL); 
  BS_ASSERT(pAccumulator != NULL); 
  
  pListBlob = (bs_blob_container_t*) pUserData; 
  pCreateData = (bs_blob_collective_create_new_joints_t*) pAccumulator; 
  
  pListBlobPointmass = bs_blob_get_middle(pListBlob->pBlob); 
  pNewBlobPointmass = bs_blob_get_middle(pCreateData->pNewBlobContainer->pBlob); 
 
  pNewJoint = bs_joint_create(pListBlobPointmass, pNewBlobPointmass, 0.0f, 0.0f); 
  dist = bs_blob_get_radius(pListBlob->pBlob) + 
    bs_blob_get_radius(pCreateData->pNewBlobContainer->pBlob); 
  bs_joint_set_consts(pNewJoint, dist, 0.0f); 
  
  pJointContainer = bs_joint_container_create(pNewJoint, 
    pCreateData->pNewBlobContainer->id, pListBlob->id);
    
  bs_list_add_element_to_front(pCreateData->pCollective->pJoints, 
    (void*) pJointContainer, bs_joint_container_destroy);   
}

void 
bs_blob_collective_split(bs_blob_collective_t *pCollective)
{
  bs_blob_collective_find_largest_t findData; 
  bs_blob_container_t *pNewBlobContainer; 
  bs_blob_t *pNewBlob; 
  bs_blob_collective_create_new_joints_t createData; 
  
  BS_ASSERT(pCollective != NULL);

  findData.maxRadius = -1.0f; 
  findData.pMotherBlob = NULL; 
  
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_find_largest_cb, &findData); 
  
  BS_ASSERT(findData.pMotherBlob != NULL); 
  
  bs_blob_scale(findData.pMotherBlob, 0.75f); 
  
  /* 0.825 = (1.0f - 0.75f) * 0.5f + 0.75f */
  bs_list_map_fp(pCollective->pJoints, bs_blob_collective_scale_old_joints, 0.825f, &findData);  
 
  pNewBlob = bs_blob_create(bs_blob_get_x_pos(findData.pMotherBlob), 
    bs_blob_get_y_pos(findData.pMotherBlob), bs_blob_get_radius(findData.pMotherBlob)); 
  pNewBlobContainer = bs_blob_container_create(pNewBlob); 

  createData.pCollective = pCollective; 
  createData.pNewBlobContainer = pNewBlobContainer; 
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_create_new_joints_cb, &createData); 
  
  bs_list_add_element_to_front(pCollective->pBlobs, pNewBlobContainer, 
    bs_blob_container_destructor); 
}

typedef struct bs_blob_collective_join_find_smallest_st
{
  bs_blob_container_t *pSmallest; 
  bs_float_t minRadius; 
} bs_blob_collective_join_find_smallest_t; 

static void 
bs_blob_collective_join_find_smallest_cb(void *pUserData, 
                                         void *pAccumulator)
{
  bs_blob_container_t *pListBlobContainer; 
  bs_blob_collective_join_find_smallest_t *pFindData; 
  bs_float_t radius; 
  
  pListBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pListBlobContainer != NULL);
  pFindData = (bs_blob_collective_join_find_smallest_t*) pAccumulator; 
  BS_ASSERT(pFindData != NULL);

  radius = bs_blob_get_radius(pListBlobContainer->pBlob); 
  if(radius < pFindData->minRadius)
  {
    pFindData->minRadius = radius; 
    pFindData->pSmallest = pListBlobContainer; 
  }    
}

typedef struct bs_blob_collective_join_find_closest_st
{
  bs_blob_container_t *pClosest; 
  bs_float_t minDist; 
  bs_pointmass_t *pBlobMiddle; 
  bs_int32_t smallestId; 
} bs_blob_collective_join_find_closest_t; 

static void 
bs_blob_collective_join_find_closest_cb(void *pUserData, 
                                        void *pAccumulator)
{
  bs_blob_container_t *pListBlobContainer; 
  bs_blob_collective_join_find_closest_t *pFindData; 
  bs_pointmass_t *pListPointmass; 
  bs_float_t aXbX, aYbY, dist; 
  
  pListBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pListBlobContainer != NULL);
  pFindData = (bs_blob_collective_join_find_closest_t*) pAccumulator; 
  BS_ASSERT(pFindData != NULL);
  
  if(pListBlobContainer->id == pFindData->smallestId)
  {
    return; 
  }
  
  pListPointmass = bs_blob_get_middle(pListBlobContainer->pBlob); 
  
  aXbX = bs_pointmass_get_x_pos(pListPointmass) - bs_pointmass_get_x_pos(pFindData->pBlobMiddle); 
  aYbY = bs_pointmass_get_y_pos(pListPointmass) - bs_pointmass_get_y_pos(pFindData->pBlobMiddle); 
  
  dist = aXbX * aXbX + aYbY * aYbY; 
  
  if(dist < pFindData->minDist)
  {
    pFindData->minDist = dist; 
    pFindData->pClosest = pListBlobContainer; 
  }    
}

static bs_bool_t
bs_blob_collective_remove_joints_cb(void *pUserData, 
                                    bs_int32_t blobId)
{
  bs_joint_container_t *pJointContainer; 
  
  pJointContainer = (bs_joint_container_t*) pUserData; 
  BS_ASSERT(pJointContainer != NULL); 
  
  if(pJointContainer->idA == blobId || 
     pJointContainer->idB == blobId)
  {
    return TRUE; 
  }    
  return FALSE; 
}

static bs_bool_t 
bs_blob_collective_remove_blob_cb(void *pUserData, 
                                  bs_int32_t blobId, 
                                  bs_bool_t *pCont)
{
  bs_blob_container_t *pBlobContainer; 

  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  if(pBlobContainer->id == blobId)
  {
    *pCont = FALSE; 
    return TRUE; 
  }
  
  return FALSE; 
}

void 
bs_blob_collective_join(bs_blob_collective_t *pCollective)
{
  bs_blob_collective_join_find_smallest_t findSmallest; 
  bs_blob_collective_join_find_closest_t findClosest; 
  bs_blob_collective_find_largest_t scaleData; 
  bs_float_t r1, r2, r3; 
  
  BS_ASSERT(pCollective != NULL);  
  BS_ASSERT(bs_list_get_length(pCollective->pBlobs) > 0);   

  if(bs_list_get_length(pCollective->pBlobs) == 1)
  {
    return; 
  }
  
  findSmallest.pSmallest = NULL; 
  findSmallest.minRadius = 10000.0f; 
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_join_find_smallest_cb, &findSmallest); 
  BS_ASSERT(findSmallest.pSmallest != NULL); 
  
  findClosest.pClosest = NULL; 
  findClosest.minDist = 10000.0f; 
  findClosest.pBlobMiddle = bs_blob_get_middle(findSmallest.pSmallest->pBlob); 
  findClosest.smallestId = findSmallest.pSmallest->id;
  bs_list_map_acc(pCollective->pBlobs, bs_blob_collective_join_find_closest_cb, &findClosest); 
  BS_ASSERT(findClosest.pClosest != NULL); 
 
  r1 = bs_blob_get_radius(findSmallest.pSmallest->pBlob); 
  r2 = bs_blob_get_radius(findClosest.pClosest->pBlob); 
  r3 = sqrt(r1 * r1 + r2 * r2); 
  bs_blob_scale(findClosest.pClosest->pBlob, 0.945f * r3 / r2); 
  
  bs_list_unlink_i(pCollective->pJoints, bs_blob_collective_remove_joints_cb, findSmallest.pSmallest->id);
  bs_list_unlink_i_cont(pCollective->pBlobs, bs_blob_collective_remove_blob_cb, findSmallest.pSmallest->id); 
  
  scaleData.id = findClosest.pClosest->id; 
  bs_list_map_fp(pCollective->pJoints, bs_blob_collective_scale_old_joints, 1.0f / 0.825f, &scaleData);  
}

static void 
bs_blob_collective_move_cb(void *pUserData, 
                           bs_float_t dt)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 

  bs_blob_move(pBlobContainer->pBlob, dt); 
}

void 
bs_blob_collective_move(bs_blob_collective_t *pCollective, 
                        bs_float_t dt)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_f(pCollective->pBlobs, bs_blob_collective_move_cb, dt);
}

static void 
bs_blob_colelctive_blob_sc_cb(void *pUserData, 
                              void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_sc(pBlobContainer->pBlob, (bs_env_t*) pDataArg); 
}

static void 
bs_blob_collective_joint_sc_cb(void *pUserData)
{
  bs_joint_container_t *pJointContainer; 

  pJointContainer = (bs_joint_container_t*) pUserData;
  BS_ASSERT(pJointContainer != NULL);

  bs_joint_sc(pJointContainer->pJoint);   
}

void 
bs_blob_collective_sc(bs_blob_collective_t *pCollective, 
                      bs_env_t *pEnv)
{
  bs_int32_t i; 
  
  BS_ASSERT(pCollective != NULL); 

  bs_list_map_p(pCollective->pBlobs, bs_blob_colelctive_blob_sc_cb, (void*) pEnv); 
  
  for(i = 0; i < 1; i++)
  {
    bs_list_map(pCollective->pJoints, bs_blob_collective_joint_sc_cb); 
  }
}

static void 
bs_blob_collective_set_force_cb(void *pUserData, 
                                void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_set_force(pBlobContainer->pBlob, (bs_vector_t*) pDataArg); 
}

void 
bs_blob_collective_set_force(bs_blob_collective_t *pCollective, 
                             bs_vector_t *pForce)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_set_force_cb, (void*) pForce); 
}

static void 
bs_blob_collective_add_force_cb(void *pUserData, 
                                void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_add_force(pBlobContainer->pBlob, (bs_vector_t*) pDataArg); 
}

void 
bs_blob_collective_add_force(bs_blob_collective_t *pCollective, 
                             bs_vector_t *pForce)
{
  BS_ASSERT(pCollective != NULL); 
  bs_list_map_p(pCollective->pBlobs, bs_blob_collective_add_force_cb, (void*) pForce); 
}

static void 
bs_blob_collective_draw_cb(void *pUserData, 
                           bs_float_t scaleFactor, 
                           void *pDataArg)
{
  bs_blob_container_t *pBlobContainer; 
  
  pBlobContainer = (bs_blob_container_t*) pUserData; 
  BS_ASSERT(pBlobContainer != NULL); 
  
  bs_blob_draw(pBlobContainer->pBlob, (cairo_t*) pDataArg, scaleFactor); 
}

void 
bs_blob_collective_draw(bs_blob_collective_t *pCollective, 
                        cairo_t *pCairo, 
                        bs_float_t scaleFactor)
{
  bs_list_map_fp(pCollective->pBlobs, bs_blob_collective_draw_cb, scaleFactor, (void*) pCairo);  
}
