#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <SDL.h>

#include "bs_cairo.h"
#include "bs_blob_collective.h"
#include "bs_assert.h"
#include "bs_mem.h"
#include "bs_list.h"

typedef struct bs_main_data_st 
{
  bs_cairo_sdl_t *pCairoSdl; 
  bs_blob_collective_t *pCollective;   
  bs_list_factory_t *pListFactory; 
  bs_env_t *pEnv; 
  bs_vector_t *pGravity; 
  double width, height; 
  bs_uint32_t newTimerInterval; 
  bs_float_t fps; 
  bs_int32_t lastFrameTimeStamp; 
  bs_bool_t running; 
} bs_main_data_t; 

static void 
bs_main_clear(cairo_t *pCairo, 
              double width, 
              double height)
{
  cairo_set_source_rgb(pCairo, 1.0, 1.0, 1.0);

  cairo_new_path(pCairo); 
  cairo_move_to(pCairo, 0.0, 0.0); 
  cairo_line_to(pCairo, width, 0.0); 
  cairo_line_to(pCairo, width, height); 
  cairo_line_to(pCairo, 0.0, height); 
  cairo_close_path(pCairo); 
  
  cairo_fill(pCairo); 
}

static void 
bs_main_draw_fps(cairo_t *pCairo, 
                 bs_main_data_t *pMainData)
{
  char fpsBuffer[32]; 
  struct timeval timeStamp; 
  bs_int32_t thisFrameTimeStamp; 
  
  gettimeofday(&timeStamp, NULL); 
  thisFrameTimeStamp = timeStamp.tv_sec * 1000 + timeStamp.tv_usec / 1000; 
  
  pMainData->fps = pMainData->fps * 0.9f + 
    0.1f * 1000.0f / (thisFrameTimeStamp - pMainData->lastFrameTimeStamp + 1.0f); 
  sprintf(fpsBuffer, "FPS: %.2f", pMainData->fps); 
  pMainData->lastFrameTimeStamp = thisFrameTimeStamp; 
  
  cairo_set_source_rgb(pCairo, 0.0, 0.0, 0.0);

  cairo_move_to(pCairo, 10.0, 20.0); 
  cairo_show_text(pCairo, fpsBuffer); 
}

static void 
bs_main_redraw(bs_main_data_t *pMainData)
{
  cairo_t *pCairo; 
  
  pCairo = bs_cairo_sdl_get_cairo(pMainData->pCairoSdl); 

  bs_cairo_sdl_lock_surface(pMainData->pCairoSdl); 
  
  bs_main_clear(pCairo, pMainData->width, pMainData->height); 
  bs_blob_collective_draw(pMainData->pCollective, pCairo, 200.0);
  bs_main_draw_fps(pCairo, pMainData); 
  SDL_UpdateRect(pMainData->pCairoSdl->pScreen, 0, 0, 0, 0);

  bs_cairo_sdl_unlock_surface(pMainData->pCairoSdl); 
}

static void 
bs_main_update_simulation(bs_main_data_t *pMainData)
{
  bs_blob_collective_move(pMainData->pCollective, 0.05); 
  bs_blob_collective_sc(pMainData->pCollective, pMainData->pEnv); 
  bs_blob_collective_set_force(pMainData->pCollective, pMainData->pGravity);     
}

static bs_uint32_t 
bs_main_sdl_timer_callback(bs_uint32_t timerInterval, 
                           void *pUserData)
{
  SDL_Event event;
  SDL_UserEvent userevent;
  bs_main_data_t *pMainData; 
  
  pMainData = (bs_main_data_t*) pUserData; 
  
  userevent.type = SDL_USEREVENT;
  userevent.code = 1;
  userevent.data1 = NULL;
  userevent.data2 = NULL;

  event.type = SDL_USEREVENT;
  event.user = userevent;

  SDL_PushEvent(&event);

  return pMainData->newTimerInterval;
}

int 
main(int argc, 
     char *argv[])
{
  bs_main_data_t *pMainData; 
  SDL_Event event; 
  bs_vector_t force; 
  int usedTime; 
  struct timeval startTime, stopTime; 
  
  BS_MEM_ALLOC_TYPE(pMainData, bs_main_data_t, 1); 
  
  pMainData->pListFactory = bs_list_factory_create(); 
  pMainData->pCollective = bs_blob_collective_create(1.0f, 1.0f, pMainData->pListFactory); 
  pMainData->pEnv = bs_env_create(0.0f, 0.0f, 3.0f, 1.8f); 
  pMainData->pGravity = bs_vector_create(0.0f, 10.0f); 
  pMainData->pCairoSdl = bs_cairo_sdl_create(600, 400); 
  pMainData->width = 600.0; 
  pMainData->height = 400.0; 
  pMainData->newTimerInterval = 50; 
  pMainData->fps = 20.0; 
  pMainData->running = TRUE; 
  
  gettimeofday(&startTime, NULL); 
  pMainData->lastFrameTimeStamp = startTime.tv_sec * 1000 + startTime.tv_usec / 1000; 
  
  SDL_AddTimer(50, bs_main_sdl_timer_callback, pMainData); 
  
  for(;;)
  {
    SDL_WaitEvent(&event); 
    
    switch(event.type)
    {
      case SDL_KEYDOWN: 
        switch(event.key.keysym.sym)
        {
          case SDLK_q: 
          case SDLK_ESCAPE: 
            exit(0); 
            break; 
      
          case SDLK_h: 
            bs_blob_collective_split(pMainData->pCollective); 
            break; 
      
          case SDLK_j: 
            bs_blob_collective_join(pMainData->pCollective); 
            break; 
      
          case SDLK_g: 
            if(bs_vector_get_y(pMainData->pGravity) > 0.0)
            {
              bs_vector_set_y(pMainData->pGravity, 0.0); 
            }
            else 
            {
              bs_vector_set_y(pMainData->pGravity, 10.0);           
            }        
            break; 
        
          case SDLK_LEFT:
            bs_vector_init(&force, -50.0, 0.0); 
            bs_blob_collective_set_force(pMainData->pCollective, &force);   
            break;      
      
          case SDLK_RIGHT: 
            bs_vector_init(&force,  50.0, 0.0); 
            bs_blob_collective_set_force(pMainData->pCollective, &force);   
            break;         
      
          case SDLK_UP: 
            bs_vector_init(&force,  0.0, -50.0); 
            bs_blob_collective_set_force(pMainData->pCollective, &force);   
            break;         
      
          case SDLK_DOWN: 
            bs_vector_init(&force,  0.0, 50.0); 
            bs_blob_collective_set_force(pMainData->pCollective, &force);   
            break;
          
          case SDLK_p:
            if(pMainData->running == TRUE)
            {
              pMainData->running = FALSE; 
              pMainData->newTimerInterval = 0;
            }
            else 
            {
              pMainData->running = TRUE; 
              pMainData->newTimerInterval = 50;
              SDL_AddTimer(50, bs_main_sdl_timer_callback, pMainData); 
            }
            break; 
          
          default: 
            break; 
        }
        break; 
      
      case SDL_USEREVENT:
        if(event.user.code == 1)
        {
          gettimeofday(&startTime, NULL); 
          bs_main_redraw(pMainData); 
          bs_main_update_simulation(pMainData);
          gettimeofday(&stopTime, NULL); 
          
          usedTime  = (stopTime.tv_sec - startTime.tv_sec) * 1000; 
          usedTime += (stopTime.tv_usec - startTime.tv_usec) / 1000; 
          
          if(usedTime < 50)
          {
            usedTime = 50; 
          }
            
          pMainData->newTimerInterval = usedTime;                        
        }          
        break; 
      
      case SDL_QUIT: 
        exit(0); 
        break; 
      
      default: 
        break; 
    }
  }
  
  return 0; 
}
