#include <stdio.h>
#include <gdk/gdkx.h>

#include "bs_cairo.h"
#include "bs_blob_collective.h"
#include "bs_assert.h"
#include "bs_mem.h"
#include "bs_list.h"

typedef struct bs_main_data_st 
{
  bs_cairo_gtk_t *pCairoGtk; 
  bs_blob_collective_t *pCollective;   
  bs_list_factory_t *pListFactory; 
  bs_env_t *pEnv; 
  bs_vector_t *pGravity; 
} bs_main_data_t; 

static void 
bs_main_redraw(void *pUserData)
{
  bs_main_data_t *pMainData; 
  cairo_t *pCairo; 
  
  pMainData = (bs_main_data_t*) pUserData;
  BS_ASSERT(pMainData != NULL); 
  BS_ASSERT(pMainData->pCollective != NULL);
  BS_ASSERT(pMainData->pCairoGtk != NULL); 
  
  pCairo = bs_cairo_gtk_get_cairo(pMainData->pCairoGtk); 
  bs_blob_collective_draw(pMainData->pCollective, pCairo, 200.0);   
  cairo_destroy(pCairo);
}

static gboolean
bs_main_timeout(void *pUserData)
{  
  bs_main_data_t *pMainData; 
  
  pMainData = (bs_main_data_t*) pUserData; 
  BS_ASSERT(pMainData != NULL); 
  BS_ASSERT(pMainData->pCollective != NULL); 
  BS_ASSERT(pMainData->pCairoGtk != NULL); 
   
  bs_blob_collective_move(pMainData->pCollective, 0.05); 
  bs_blob_collective_sc(pMainData->pCollective, pMainData->pEnv); 
  bs_blob_collective_set_force(pMainData->pCollective, pMainData->pGravity);   
  
  gtk_widget_queue_draw(pMainData->pCairoGtk->pWindow); 
  
  return TRUE; 
}

static gboolean 
bs_main_on_key_press(GtkWidget *pWidget,
                     GdkEventKey *pKey,
                     gpointer userData)
{
  bs_vector_t force; 
  bs_main_data_t *pMainData; 
  
  pMainData = (bs_main_data_t*) userData; 
  BS_ASSERT(pMainData != NULL); 
  BS_ASSERT(pMainData->pCollective != NULL); 
  BS_ASSERT(pMainData->pGravity != NULL); 
  
  if (pKey->type == GDK_KEY_PRESS)
  {
    printf("key: %d\n", pKey->keyval); 
    switch (pKey->keyval)
    {
      case 113: // 'q' quit 
      case 65307: // esc key
        gtk_main_quit ();
        break;
      case 103: // 'g' gravity
        if(bs_vector_get_y(pMainData->pGravity) > 0.0)
        {
          bs_vector_set_y(pMainData->pGravity, 0.0); 
        }
        else 
        {
          bs_vector_set_y(pMainData->pGravity, 10.0);           
        }
        break; 
      case 104: // 'h' split 
        bs_blob_collective_split(pMainData->pCollective); 
        break; 
      case 106: // 'j' split 
        bs_blob_collective_join(pMainData->pCollective); 
        break; 
      case 65361: // left arrow
        bs_vector_init(&force, -50.0, 0.0); 
        bs_blob_collective_set_force(pMainData->pCollective, &force);   
        break; 
      case 65363: // right arrow
        bs_vector_init(&force,  50.0, 0.0); 
        bs_blob_collective_set_force(pMainData->pCollective, &force);   
        break; 
      case 65362: // up arrow
        bs_vector_init(&force, 0.0, -50.0); 
        bs_blob_collective_set_force(pMainData->pCollective, &force);   
        break; 
      case 65364: // down arrow
        bs_vector_init(&force, 0.0,  50.0); 
        bs_blob_collective_set_force(pMainData->pCollective, &force);   
        break; 
      default:
        break;       
    }
  }

  return FALSE;
}

int 
main(int argc, 
     char *argv[])
{
  bs_main_data_t *pMainData; 
  
  BS_MEM_ALLOC_TYPE(pMainData, bs_main_data_t, 1); 
  
  pMainData->pListFactory = bs_list_factory_create(); 
  pMainData->pCollective = bs_blob_collective_create(1.0f, 1.0f, pMainData->pListFactory); 
  pMainData->pEnv = bs_env_create(0.0f, 0.0f, 3.0f, 2.0f); 
  pMainData->pGravity = bs_vector_create(0.0f, 10.0f); 

  gtk_init(&argc, &argv); 
  pMainData->pCairoGtk = bs_cairo_gtk_create(600, 400, bs_main_redraw, (void*)pMainData); 
  
  gtk_timeout_add(50, (GtkFunction)bs_main_timeout, (void*)pMainData); 
  g_signal_connect (G_OBJECT (pMainData->pCairoGtk->pWindow), "key-press-event", 
    G_CALLBACK (bs_main_on_key_press), (void*)pMainData); 

  gtk_main(); 
  
  bs_cairo_gtk_destroy(pMainData->pCairoGtk); 
  
  return 0; 
}
